@@ -5,7 +5,7 @@ use warnings;
 package LINQ::Collection;
 
 our $AUTHORITY = 'cpan:TOBYINK';
-our $VERSION   = '0.000_005';
+our $VERSION   = '0.000_006';
 
 use Role::Tiny;
 use LINQ::Util::Internal ();
@@ -190,77 +190,68 @@ sub take_while {
 	my $self    = shift;
 	my $filter  = LINQ::Util::Internal::assert_code( @_ );
 	my $stopped = 0;
-	my $idx     = 0;
+	my $iter    = $self->to_iterator;
 	
 	require LINQ;
-	require LINQ::Iterator;
-	'LINQ::Iterator'->new( sub {
-		local ( $@, $_ );
-		my $next;
-		my $ok = eval { $next = $self->element_at( $idx ); not $stopped };
-		my $e  = $@;
-		if ( $ok and $filter->( $_ = $next ) ) {
-			++$idx;
-			return $next;
+	LINQ::Util::Internal::create_linq( sub {
+		return LINQ::END() if $stopped;
+		my @got = $iter->();
+		if ( ! @got or ! $filter->( $_ = $got[0] ) ) {
+			$stopped++;
+			return LINQ::END();
 		}
-		if ( defined $e and not $ok ) {
-			require Scalar::Util;
-			die( $e ) unless Scalar::Util::blessed( $e );
-			die( $e ) unless $e->isa( 'LINQ::Exception::NotFound' );
-			die( $e ) unless $e->collection == $self;
-		}
-		$stopped = 1;
-		return LINQ::END();
+		return $got[0];
 	} );
 } #/ sub take_while
 
 sub skip {
 	my $self = shift;
 	my ( $n ) = @_;
-	$self->where( sub { $n-- <= 0 } );
+	$self->skip_while( sub { $n-- > 0 } );
 }
 
 sub skip_while {
-	my $self     = shift;
-	my $filter   = LINQ::Util::Internal::assert_code( @_ );
-	my $skipping = 1;
-	$self->where(
-		sub {
-			$skipping = 0
-				if $skipping
-				&& !$filter->( $_ );
-			not $skipping;
-		}
-	);
+	my $self    = shift;
+	my $filter  = LINQ::Util::Internal::assert_code( @_ );
+	my $stopped = 0;
+	my $started = 0;
+	my $iter    = $self->to_iterator;
+	
+	require LINQ;
+	LINQ::Util::Internal::create_linq( sub {
+		SKIPPING: {
+			return LINQ::END() if $stopped;
+			my @got = $iter->();
+			if ( ! @got ) {
+				$stopped++;
+				redo SKIPPING;
+			}
+			return $got[0] if $started;
+			if ( $filter->( $_ = $got[0] ) ) {
+				redo SKIPPING;
+			}
+			++$started;
+			return $got[0];
+		};
+	} );
 } #/ sub skip_while
 
 sub concat {
-	my @collections = @_;
+	my @collections = map $_->to_iterator, @_;
 	my $idx = 0;
 	
 	require LINQ;
-	require LINQ::Iterator;
-	'LINQ::Iterator'->new( sub {
-		local ( $@, $_ );
-		my $next;
+	LINQ::Util::Internal::create_linq( sub {
 		FIND_NEXT: {
 			return LINQ::END() if not @collections;
 			
-			my $ok = eval { $next = $collections[0]->element_at( $idx ); 1 };
-			my $e  = $@;
-			if ( $ok ) {
-				++$idx;
-				return $next;
+			my @got = $collections[0]->();
+			if ( not @got ) {
+				shift @collections;
+				redo FIND_NEXT;
 			}
-			elsif ( defined $e and not $ok ) {
-				require Scalar::Util;
-				die( $e ) unless Scalar::Util::blessed( $e );
-				die( $e ) unless $e->isa( 'LINQ::Exception::NotFound' );
-				die( $e ) unless $e->collection == $collections[0];
-			}
-			shift @collections;
-			$idx = 0;
-			redo FIND_NEXT;
+			
+			return $got[0];
 		};
 	} );
 }
@@ -605,15 +596,17 @@ sub zip {
 	my $other = shift;
 	my $map   = LINQ::Util::Internal::assert_code( @_ );
 	
-	my @self  = $self->to_list;
-	my @other = $other->to_list;
+	my $iter1 = $self->to_iterator;
+	my $iter2 = $other->to_iterator;
 	my @results;
 	
-	while ( @self and @other ) {
-		push @results, scalar $map->( shift( @self ), shift( @other ) );
-	}
-	
-	LINQ::Util::Internal::create_linq( \@results );
+	require LINQ;
+	LINQ::Util::Internal::create_linq( sub {
+		my @r1 = $iter1->();
+		my @r2 = $iter2->();
+		return LINQ::END unless @r1 && @r2;
+		$map->( $r1[0], $r2[0] );
+	} );
 } #/ sub zip
 
 sub default_if_empty {
@@ -943,6 +936,9 @@ Takes just the first N items from a collection, returning a new collection.
 Takes items from the collection, stopping at the first item where CALLABLE
 returns false.
 
+If CALLABLE dies, there are some issues on older versions of Perl with the
+error message getting lost.
+
 =item C<< skip( N ) >>
 
 Skips the first N items from a collection, and returns the rest as a new
